/*
 * linux/drvers/char/lpcid.c - kernel driver for caller id parallel port device
 *
 * Copyright (C) 1998  Juha Tuomala
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * <Juha.Tuomala@iki.fi> / Sateentie 4 A 32, 02100 ESPOO FINLAND
 *
 * Compile module with 'gcc -O2 -DMODULE -D__KERNEL__ -c lpcid.c'
 */

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
MODULE_LICENSE("GPL");
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include "lpcid.h"

#define VERSION "0.4.0"
#define LP_NO 3

#undef LPCIDDEBUG

struct lpcid_struct lpcid_table[] = {
  { 0x3bc, 0, {}, 0, 0 },
  { 0x378, 7, {}, 0, 0 },
  { 0x278, 0, {}, 0, 0 },
};


int init_module(void)
{
  printk("Initializing lpcid.\n");
  return lpcid_init();
}


void cleanup_module(void)
{
  printk("Unloading lpcid.\n");
  unregister_chrdev(LPCID_MAJOR, "lpcid");
}


static ssize_t lpcid_read(struct file *file, char *buf,
                          size_t count, loff_t *ppos)
{
  unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);

  int temp;

#ifdef LPCIDDEBUG
  printk("lpcid_read called\n");
#endif

  if (!LPCID_CH_RECEIVED(minor)) {
    interruptible_sleep_on(&LPCID_WAIT_Q(minor));
  }

  temp = LPCID_CH(minor);

#ifdef LPCIDDEBUG
  printk("lpcid-read: Char was: 0x%x = %d\n", temp, temp);
#endif

  memcpy(buf, &LPCID_CH(minor), 1);
  LPCID_CH_RECEIVED(minor) = 0;

  return 1;
}

static unsigned int lpcid_poll(struct file *file, poll_table *wait)
{
  unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);

#ifdef LPCIDDEBUG
  printk("lpcid_poll called\n");
#endif

  poll_wait(file, &LPCID_WAIT_Q(minor), wait);
  if (LPCID_CH_RECEIVED(minor))
    return POLLIN | POLLRDNORM;

  return 0;
}


static void lpcid_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  int minor = -1;
  int i;

  /* How should this be done, I mean _really_? */
  for (i = 0; i < LP_NO; ++i) {
    if (irq == LPCID_IRQ(i)) {
      minor = i;
      break;
    }
  }
  if (minor == -1)
    return;

#ifdef LPCIDDEBUG
  printk("!!!!!!!!!!! lpcid interrupt !!!!!!!!!!\n");
#endif

  wake_up_interruptible(&LPCID_WAIT_Q(minor));
  LPCID_CH_RECEIVED(minor) = 1;
  LPCID_CH(minor) = inb(LPCID_STATUS(minor)) & 0xb8;
}


static int lpcid_open(struct inode *inode, struct file *file)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int retval;
  unsigned int irq;

#ifdef LPCIDDEBUG
  printk("lpcid_open called\n");
#endif

  if (minor >= LP_NO)
    return -ENODEV;

  irq = LPCID_IRQ(minor);
  retval = request_irq(irq, lpcid_interrupt, SA_INTERRUPT, "lpcid", NULL);

#ifdef LPCIDDEBUG
  printk("lpcid-open: request_irq returns %d \n", retval);
#endif

  if (retval) {
    /* kfree_s(lpcid_table[minor].lpcid_buffer, LPCID_BUFFER_SIZE); */

    LPCID_CH_RECEIVED(minor) = 0;

    printk("lpcid%d unable to use interrupt %d, error %d\n", minor, irq, retval);

    return -EBUSY;
  } else {
    /* Enable IRQ in used parallel port */
    outb((LPCID_PSELECP|LPCID_PINITP|LPCID_PINTEN), (LPCID_CONTROL(minor)));
    outb(LPCID_VCC_ON,LPCID_BASE(minor));
    sti();

    MOD_INC_USE_COUNT;
    return 0;
  }
}


static int lpcid_release(struct inode *inode, struct file *file)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int irq;

#ifdef LPCIDDEBUG
  printk("lpcid_release called.\n");
#endif

  irq = LPCID_IRQ(minor);
  if (irq) {
    free_irq(irq, NULL);
    /* kfree_s(lpcid_table[minor].lpcid_buffer, LPCID_BUFFER_SIZE); no malloc used currently */
    outb(LPCID_VCC_OFF, LPCID_CONTROL(minor));
    MOD_DEC_USE_COUNT;
  }
  return 0;
}


static int lpcid_ioctl(struct inode *inode, struct file *file,
                       unsigned int cmd, unsigned long arg)
{
  unsigned int minor = MINOR(inode->i_rdev);
  int retval = 0;

#ifdef LPCIDDEBUG
  printk(KERN_DEBUG "lpcid%d ioctl, cmd: 0x%x, arg: 0x%x\n", minor, cmd, arg);
#endif

  if (minor >= LP_NO)
    return -ENODEV;

  switch (cmd) {
  case LPCIDSETDEBUG:
    printk("lpcid-ioctl: Debugging turned ON.\n");
/*    lpciddebug=1; */
    break;
  case LPCIDUNSETDEBUG:
    printk("lpcid-ioctl: Debugging turned OFF.\n");
/*    lpciddebug=0; */
    break;

  case LPCIDGETIRQ:
    retval = verify_area(VERIFY_WRITE, (void *)arg, sizeof(int));
    if (retval)
      return retval;

    memcpy((int *)arg, &LPCID_IRQ(minor), sizeof(int));
    break;

  case LPCIDSETIRQ:
    break;

  default:
    retval = -EINVAL;
  }

  return retval;
}


static struct file_operations lpcid_fops = {
  read:    lpcid_read,
  poll:    lpcid_poll,
  ioctl:   lpcid_ioctl,
  open:    lpcid_open,
  release: lpcid_release,
};


int lpcid_probe(int offset)
{
  /* Currently no known hardware support autoprobing */

  if (lpcid_table[offset].irq > 0) {
    printk("    Cid device driver set for lp%d in 0x%x\n", offset, lpcid_table[offset].base);

    /* Enable IRQ in tested parallel port */
    /* outb((LPCID_PSELECP|LPCID_PINITP|LPCID_PINTEN), (LPCID_C(offset))); */

    return 0;
  }

  return -1;
}


/* lpcid_init(), called by mem.c. Initializes lpcid and registers
 * char_dev.
 */
int lpcid_init(void)
{
  int offset;
  int count=0;
  int i;

  printk("lpcid: CallerID driver by Juha.Tuomala@iki.fi and joel@rosdahl.net.  Version %s-ALPHA.\n", VERSION);

  for (i = 0; i < 3; i++) {
    init_waitqueue_head(&lpcid_table[i].wait_q);
  }

  if (register_chrdev(LPCID_MAJOR,"lpcid", &lpcid_fops)) {
    printk("lpcid-init: Unable to get major %d\n", LPCID_MAJOR);
    return -EIO;
  }

  for (offset = 0; offset < LP_NO; offset++) {
    if (!lpcid_probe(offset)) {
      count++;
    }
  }

  if (!count) {
    printk("Unable to find any hardware. Check source file for irqs\n");
  }

  return 0;
}
